#!/usr/bin/perl -w


package DAASIlib::SMTP;

use strict;
use warnings;
use vars qw($VERSION);
our $VERSION = "0.1";

use Log::Log4perl qw(:levels);
my $logger = Log::Log4perl->get_logger(''); 
use DAASIlib::Data;
my $data = new DAASIlib::Data;


my $enhancedheader = "User-Agent: Net::SMTP\nMIME-Version: 1.0\nContent-Type: text/plain; charset=utf-8\nContent-Transfer-Encoding: 8bit\n";

my $defaultagent = "Net::SMTP";
my $defaultcontenttype = "text/plain";
my $boundary = "frontier";
my $defaultmulticontenttype = "multipart/mixed; boundary=\"$boundary\"";
my $defaulthtmlmulticontenttype = "multipart/alternative; boundary=\"$boundary\"";
my $defaultcharset = "utf-8";
my $defaultencoding = "8bit";

my $enhancedmailxheader = ' -a "User-Agent: Net::SMTP" -a "MIME-Version: 1.0" -a "Content-Type: text/plain; charset=utf-8" -a "Content-Transfer-Encoding: 8bit"';

use Net::SMTP;

sub new {
    my $class;
    my $self;

    my ( $this , $param ) = @_;
    $class = ref($this) || $this;

    $self={
	data => {},
    };

    bless($self , $class );

    return $self;

}

sub set_server {
    my $self = shift;
    my ($server, $charset, $encoding, $agent, $hello ) = @_;

    $self->{data}{smtprelay} = $server;
    if ( $server eq 'LOCAL' | $server eq 'SYSTEM' ) {
	$self->{data}{norelay} = 1;
    }

    if ( $hello ) {
	$self->{data}{hello} = $hello;
    }

    if ( $charset ) {
	$self->{data}{charset} = $charset;
    }
    else {
	$self->{data}{charset} = $defaultcharset;
    }

    if ( $encoding ) {
	$self->{data}{encoding} = $encoding;
    }
    else {
	$self->{data}{encoding} = $defaultencoding;
    }

    if ( $agent ) {
	$self->{data}{agent} = $agent;
    }
    else {
	$self->{data}{agent} = $defaultagent;
    }

}

sub send_mail {
   my $self = shift;
   my ($from, $subject, $mail_data, 
       $to_mail_addresses, $cc_mail_addresses, 
       $bcc_mail_addresses, $authname, $ishtml, $attachments ) = @_;
   my $smtp;

   if ( ! $self->{data}{smtprelay} ) {
       $logger->error("send_mail() called without a server set");
       $logger->error("you have to first call set_server, before you can use method send_mail");
       print "you have to first call set_server, before you can use method send_mail\n";
       exit();
   }

   my $contenttype;
   
   if ( $attachments ) {
       $contenttype = $defaultmulticontenttype;
   }
   elsif ( $ishtml )  {
       $contenttype = $defaulthtmlmulticontenttype;
   }
   else {
       $contenttype = $defaultcontenttype;
   }

#   my $hello = $data->getip();
#   my $hello = 'usha.daasi.int';

   my $header;
   my $timestamp = $self->create_timestamp();
     

   if ( $self->{data}{norelay} ) { 
       $header = ' -a "From: '.$from.'"';
       $header .= ' -a "Date: '.$timestamp.'"';
       $header .= $enhancedmailxheader;

       my $adds = '';
       if ( $cc_mail_addresses ) {
	   $adds .= ' -c '.$cc_mail_addresses;
       }
       if ( $bcc_mail_addresses ) {
	   $adds .= ' -b '.$bcc_mail_addresses;
       }
       $mail_data = "\"$mail_data\"";

       my $command = "echo $mail_data | mailx -s \"$subject\" $adds $header $to_mail_addresses";
      $logger->error("sending the following to system: \n$command\n");  
       system ($command);
   }
   else {

       $header = "From: $from\nTo: $to_mail_addresses\n";
       $smtp = Net::SMTP->new($self->{data}{smtprelay},
			      Hello => $self->{data}{hello},
	   );
       
       if (! $smtp) {
	   $logger->error("first try to connect to relay failed");
	   $logger->error("will try again in 10 seconds");
	   sleep 10;
	   $smtp = Net::SMTP->new($self->{data}{smtprelay});
	   if (! $smtp) {
	       $logger->error("second try to connect to relay failed");
	       $logger->error("will abort sending mail");
	       print "##### ERROR couldn't open new connection to smtp server!!\n";
	       return 0;
	   }
       }
       
       my $authpw='';
       if ( $authname ) {
	   $authpw = $data->getSecret('smtpauth');
	   $smtp->auth($authname,$authpw);
	   $logger->debug("performed SMTP AUTH with user $authname");
       }
       
       $smtp->mail($from);
       $smtp->to  ( split ( /,/, $to_mail_addresses  ));

# $smtp->to  ( split ( /,/, $to_mail_addresses  ), { Notify => ['FAILURE','DELAY'], SkipBad => 1 }); 
# könnte man noch hinten dran stellen noch nicht getestet!
       
       if ( $cc_mail_addresses ) {
	   $smtp->cc  ( split ( /,/, $cc_mail_addresses  ));
	   $header .= "Cc: $cc_mail_addresses\n";
       }
       
       if ( $bcc_mail_addresses ) {
	   $smtp->bcc ( split ( /,/, $bcc_mail_addresses ));
       }
       
       my $timestamp = $self->create_timestamp();

       $header .= "Date: $timestamp\n";
       $header .= "Subject:$subject\n";
       $header .= "User-Agent: ".$self->{data}{agent}."\n";

       $header .= "MIME-Version: 1.0\n";


       $header .= "Content-Type: $contenttype";
       if ( ! $ishtml && ! $attachments ) {
	   $header .= "; charset=".$self->{data}{charset}."\n";
	   $header .= "Content-Transfer-Encoding: ".$self->{data}{encoding}."\n\n";
	   $smtp->data($header,
		       $mail_data);
       }
       else {
	   $header .= "\n\n";
	   $header .= "This is a message with multiple parts in MIME format.\n";
	   my $htmlmail_data='';
	   if ( $ishtml ) {
	       $htmlmail_data = $self->create_html_mail($mail_data); 
	       $smtp->data();
	       $smtp->datasend($header);
	       $smtp->datasend("--$boundary\n");
     	       $smtp->datasend("Content-type: $defaultcontenttype; charset="
			       .$self->{data}{charset}."\n");
#	       $smtp->datasend("Content-Disposition: quoted-printable\n");
	       $smtp->datasend("Content-Transfer-Encoding: "
			       .$self->{data}{encoding}."\n\n");
	       $smtp->datasend($mail_data);
	       $smtp->datasend("--$boundary\n");
	       $smtp->datasend("Content-type: text/html; charset="
			       .$self->{data}{charset}."\n");
	       $smtp->datasend("Content-Transfer-Encoding: "
			       .$self->{data}{encoding}."\n\n");
	       $smtp->datasend($htmlmail_data);
	       $smtp->datasend("\n--$boundary--\n");
	       $smtp->dataend();
	   }
	   else {
	       ### TODO handle attachements here
	   }
       }

       $smtp->quit;
   }
   return 1;

} # send_mail 

sub create_html_mail {
    my $self = shift;
    my ( $data ) = @_;

    my $header = '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">'
	."\n"
	.'<html><head><meta http-equiv="Content-Type" content="text/html; charset="'
	.$self->{data}{charset}. '" />'
	."\n</head>\n\n<body>";

#    print "|$header|\n";

    $data =~ s|([ \n])\*(.*?)\*([ \n.,])|$1<b>$2</b>$3|g;

    if ( $data =~ m{http://} ) {
	$data =~ s{(http://.*?)([ )\n]|[\.,;:] )}{<a href="$1">$1</a>$2}g;
	$data =~ s{(https://.*?)([ )\n]|[\.,;:] )}{<a href="$1">$1</a>$2}g;

    }
    else {
	$data =~ s{([^/])(www\..*?)([ )\n]|[\.,;:] )}{$1<a href="http://$2">$2</a>$3}g;
	$data =~ s{(https://.*?)([ )\n]|[\.,;:] )}{<a href="$1">$1</a>$2}g;
    }
    $data =~ s{([\w\-\.]*?\@[\w\-\.]*?)([ )\n]|[\.,;:] )}{<a href="mailto:$1">$1</a>$2}g;
    $data =~ s|\n|<br>\n|g;
    
    $data = "$header$data</html>"; 

    return ( $data);
}



sub create_timestamp {
    my $self = shift;
    
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();

    my $utcdiff = 2;

    if ( $isdst ) {
	$utcdiff++;
    }

    my @monabbr = qw( Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec );
    my $monstr = $monabbr[$mon];

    my @wdayabbr = qw( Sun Mon Tue Wed Thu Fri Sat );
    my $wdaystr = $wdayabbr[$wday];

    $year+=1900;
#    my $datestring = sprintf("${wdaystr}, $mday $monstr $year ${hour}:${min}:${sec} +0${utcdiff}00";
    my $datestring = sprintf("%s, %02d %s %04d %02d:%02d:%02d +%02d00", 
			     $wdaystr,$mday,$monstr,$year,
			     $hour,$min,$sec,$utcdiff);
    return $datestring;

}

1;
