<?php

class TGLdap {
  
  var $conn;
  var $conf;

  function __construct( &$conf ) {
    $this->conf =& $conf;
  
    $ldap = $conf['ldap'];
    $this->conn = ldap_connect($ldap['host'], $ldap['port'])
          or die("Keine Verbindung zum LDAP Server.");      
          
    ldap_set_option($this->conn,LDAP_OPT_PROTOCOL_VERSION,3);
    
    if($ldap['tls']) {
      ldap_start_tls( $this->conn ) or die('TLS not possible');
    }
    
    $bound = ldap_bind($this->conn, $ldap['managerdn'], $ldap['managerpassword']);
  }
  
  // return full-name of logged in user (e.g. for mail)
  function getUsername($login) {
     
     $basedn = 'ou=users,dc=textgrid,dc=de';
     $filter = '(uid='.$login.')';
     $justthese = array("cn");
     
     $result = ldap_search( $this->conn, $basedn, $filter, $justthese);
     $entry = ldap_get_entries($this->conn, $result);
     
     return $entry[0]['cn'][0];
  }
  
  function uidExists($uid) {
    // retrieving entry via complete dn and ldap_read() is no good idea, couse there may be more than one uid,
    // but only one uid is in the dn
    // $dn = 'uid='.$uid.',ou=users,dc=textgrid,dc=de';
    
    $basedn = 'ou=users,dc=textgrid,dc=de';
    $filter = '(uid='.$uid.')';
    $justthese = array("");

    $result = ldap_search( $this->conn, $basedn, $filter, $justthese);
    
    // return true if result found
    return (ldap_count_entries($this->conn, $result) > 0);
    
  }
  
  function eppnExists($eppn) {

    $basedn = 'ou=users,dc=textgrid,dc=de';
    $filter = '(eduPersonPrincipalName='.$eppn.')';
    $justthese = array("");

    $result = ldap_search( $this->conn, $basedn, $filter, $justthese);
    
    // return true if result found
    return (ldap_count_entries($this->conn, $result) > 0);
    
  } 
  
  function emailExists($email) {
    
    $basedn = 'ou=users,dc=textgrid,dc=de';
    $filter = '(mail='.$email.')';
    $justthese = array("");

    $result = ldap_search( $this->conn, $basedn, $filter, $justthese);
    
    // return true if result found
    return (ldap_count_entries($this->conn, $result) > 0);
  }
  
  /**
   * TODO: userObject would be REALLY nice
   */
  function addUser($data) {
    
    $uid = $data['uid'];
    
    //objectclass presets
    $in['objectclass'][0] = 'top';
    $in['objectclass'][1] = 'inetOrgPerson';
    $in['objectclass'][2] = 'eduPerson';
    $in['objectclass'][3] = 'TextGridUser';
    
    // Organisation, optional
    if($data['o'])
      $in['o'] = $data['o'];
    
    // Vorname
    $in['givenName'] = $data['givenName'];
    // Nachname
    $in['sn'] = $data['sn'];

    // http://www.openldap.org/faq/data/cache/347.html
    $in['userPassword'] = "{SHA}" . base64_encode(sha1( $data['password'], TRUE )); 
    
    $in['mail'] = $data['mail'];
    
    $in['eduPersonPrincipalName'] = $data['eduPersonPrincipalName'];
    
    $in['uid'] = $uid;
    $in['cn'] =  $data['cn'];
    
    $in['TGWantsNewsletter'] = $data['TGWantsNewsletter'];
    
    $dn = 'uid='.$uid.',ou=users,dc=textgrid,dc=de';
    $r = ldap_add($this->conn, $dn, $in);
    
    return $r;
  } 
  
  // add user (eppn) to textgridproject with given id
  function addUserToProject($eppn, $projectid) {
    
    $in["rbacPerformer"] = $eppn;
    
    $dn = 'rbacName=Bearbeiter,rbacName='.$projectid.',rbacName=Projekt-Teilnehmer,ou=roles,ou=rbac,dc=textgrid,dc=de';
    $r = ldap_mod_add($this->conn, $dn, $in);
    
    return $r;
  }
  
}
?>
