#!/usr/bin/perl -w
######################################################################
# TextGrid demo portal used for delegation of short-lived credentials
# initially based on gridshib-ca-demo-portal.cgi.in,v 1.5 2006/11/20 00:07:38 welch Exp 
# 
# (c) Martin Haase / DAASI International GmbH / Gap-SLC
# Version 2.0 
# Fr 23. Jul 09:34:20 CEST 2010
######################################################################

use CGI;
use XML::Simple;

my $conf = "../../../config_tgwebauth.xml";

######################################################################
$cgi = new CGI;
$status = $cgi->param("status");

if (!defined($status))
{
    # No status, we are being called for the first time by the user
    # Create a certificate request and create a form for requesting
    # delegation.
    makeRequest();
}
elsif ($status eq "success")
{
    # Successful delegation from GridShib-CA. Read in certificate
    # and store.
    handleSuccess();
}
elsif ($status eq "rejected")
{
    # We were rejected for some reason. Clean up.
    handleRejected();
}

exit(0);

######################################################################

sub makeRequest
{
    use MIME::Base64;

    $userDetails = $cgi->param("userDetails");
    if (not defined ($userDetails)) {
	errorExit("Sorry, your Browser is not supported.");
    }
    @all = split "\\|", $userDetails;
    $eppn = $all[0];
    $sid = $all[1];
    $authZinstance = $all[2];

    $slcConfig = getConfig($conf, $authZinstance);

    # generate certificate request with key remotely
    $rethash = invokeWebService (
	$slcConfig->{rbacbase}.'tgextra.php', 
	'http://textgrid.info/namespaces/middleware/tgauth',
	'getCSRRequest', {'auth' => $sid}
	);

    $reqPEM = decode_base64 ( $rethash->{csr} );

    my $targetURL = $slcConfig->{SLCSaddress};
    # this does not work on ws1:
    # my $myURL = $cgi->url(); 
    $myURL = $slcConfig->{portalDelegationURL};

    print $cgi->header();
    print $cgi->start_html("TextGrid / SLC Demo");
    print <<"EOF";
<center>
Welcome to the SLC demo for TextGrid for YOU ($eppn)
</center>
<p>
This script shows how the TextGrid middleware can ask for a delegated <b>S</b>hort-<b>L</b>ived <b>C</b>redential
from the DFN SLC Service by redirecting a user there with a credential request.
<p>
This demo uses the GridShib CA at: $targetURL
<p>
Your certificate request has been generated.
<p>
<form action="$targetURL" method="get">
<input name="certificateRequest" type="hidden" value="$reqPEM">
<input name="portalURL" type="hidden" value="$myURL">
<input name="portalData" type="hidden" value="$userDetails">
<input value="Click to submit Delegation request" type="submit">
</form>
EOF
    print $cgi->end_html();
    
}

sub handleSuccess
{
    my $certificate = $cgi->param("certificate");
    my $userDetails = $cgi->param("portalData");

    @all = split "\\|", $userDetails;
    $eppn = $all[0];
    $sid = $all[1];
    $authZinstance = $all[2];

    $slcConfig = getConfig($conf, $authZinstance);

    $certificate =~ s/\x0d\x0a/\x0a/g;

    $rethash = invokeWebService (
        $slcConfig->{rbacbase}.'tgextra.php',
	'http://textgrid.info/namespaces/middleware/tgauth',
	'putCRTRequest', {'auth' => $sid, 'crt' => $certificate}
	);

    $noDelegationURL = $slcConfig->{noDelegationURL};

    $success = $rethash->{success};
    if ($success eq 'false') {
        errorExitWithSID($eppn, $sid, $noDelegationURL , "Could not store your credentials in TG-auth*.");
    }

    print $cgi->header();
    print $cgi->start_html("Delegation Successful");
    print <<"EOF";
Your delegated credential was successfully retrieved and stored in TG-auth*. Here is its public key:
<p>
<pre>
$certificate
</pre>
The corresponding private key is now held by TG-auth*.
<p>
<form method="post" action="$noDelegationURL">
  <input type="hidden" name="ePPNplusSID" value="$eppn|$sid" />
  <input type="submit" value="Return to the TextGridLab"/>
</form><br/>
EOF

registerUserAtVOMRS ( 
    "/usr/local/bin/VOMRSclient/bin/runAutoregClient.sh", 
    $certificate, 
    $slcConfig->{rbacbase},
    $sid
);

print $cgi->end_html();

}

sub handleRejected
{
    print $cgi->header();
    print $cgi->start_html("Delegation Request Rejected");
    print "The delegation failed.";
    print $cgi->end_html();
}

######################################################################
#
# Utility functions
#

### we cannot use &invokeWebservice as SOAP::Lite only unwillingly treats XML attributes right
### however, the Web Service returns datastructures with XML attributes, such as:
#<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns1="http://textgrid.info/namespaces/middleware/tgauth">
#   <SOAP-ENV:Body>
#      <ns1:getMyUserAttributesResponse>
#         <attribute name="givenname" mandatory="true" displayname="Given Name">
#            <value>Martin</value>
#            <description>Your First Name</description>
#         </attribute>
#         <attribute name="surname" mandatory="true" displayname="Surname">
#            <value>Haase</value>
#            <description>Your Last Name</description>
#         </attribute>
#         ...
# The mapping onto D-Grid VOMRS attribute names from ours (e.g. 'street' => 'Street or P.O. Box') is also done here
sub getAttributes {
    my ($rbacbase, $sid) = @_;

    my ($endpointInWSDL,$namespace,$methodname,$arghash) = 
	($rbacbase . 'tgextra.php',
	 'http://textgrid.info/namespaces/middleware/tgauth',
	 'getMyUserAttributesRequest',
	 {'auth' => $sid }
	);	

    use SOAP::Lite;

    my $soap = SOAP::Lite->new( proxy => $endpointInWSDL);
    $soap->default_ns($namespace);

    @args = ();
    foreach $k (keys %{$arghash}) {
	push @args, SOAP::Data->name($k)->value( $arghash->{$k});
    }

    my $result = $soap->call( $methodname, @args );

    die $result->fault->{ faultstring } if ($result->fault);

    $i=0;
    $attrs = $result->freeform->{'attribute'};
    foreach my $a ($result->dataof("//attribute/")) {
	$attrs->[$i]->{'name'} = $a->attr->{'name'};
	$i++;
    }
    
    my %vomrsmap = ('givenname' => 'First name',
		    'surname' => 'Last name',
		    #'displayname' => '',
		    #'mail' => '',
		    'organisation' => 'Institute or Department',
		    #'orgunit' => '',
		    'street' => 'Street or P.O. Box',
		    'plz' => 'Zipcode',
		    'city' => 'City',
		    'country' => 'Country',
		    'tel' => 'Phone',
		    'citizenship' => 'Nationality',
		    #'interest' => '',
		    #'personid' => '',
		    #'agreesearch' => '',
	);

    %result = ();

    foreach $entry (@{$attrs}) {
	if (exists $vomrsmap{$entry->{'name'}}) {
            $result{$vomrsmap{$entry->{'name'}}} = $entry->{'value'};
	}
    }

    return \%result;
}


sub registerUserAtVOMRS {
    my ($registerScript, $pem, $rbacbase, $sid) = @_;

    $attrhash = getAttributes ($rbacbase, $sid);

    use Crypt::OpenSSL::X509;

    my $crt = Crypt::OpenSSL::X509->new_from_string ( $pem );

    my $dn_commas = $crt->subject();
    my $ca_commas = $crt->issuer();
    my $mail = $crt->email();
    my $serial = $crt->serial();

#    my ($firstname, $lastname ) = getSLCname ($dn_commas);

    my $dn = commas2slashes ($dn_commas);
    my $ca = commas2slashes ($ca_commas);

    $personalinfo = "";
    foreach $key (keys %{$attrhash}) {
	$personalinfo = $personalinfo . $key . ',' . $attrhash->{$key} . ',';
    }
    chop $personalinfo;

    system "$registerScript '$dn' '$ca' '$serial' '$mail' '$personalinfo' >/dev/null";
}


# assumes "C=DE, ..., CN=Tanja Test - tanja.test@textgrid-test-idp.de"
# returns ("Tanja", "Test")
sub getSLCname {
    my $dn = shift;
    $dn =~ /CN=(\S+\s+)+(\S+)\s+-\s+/;
    $givennames = $1;
    $surname = $2;
    $givennames =~ s/\s+$//;
    return ( $givennames, $surname );
}

# translates C=DE, O=DFN-Verein, OU=DFN-PKI, CN=DFN-Verein Test-AAI SLCS CA
# to /C=DE/O=DFN-Verein/OU=DFN-PKI/CN=DFN-Verein Test-AAI SLCS CA
sub commas2slashes {
    $_ = shift;
    s/,\s+/\//g;
    return '/' . $_;
}


sub getConfig {
    my ($configfilename, $authZinstance) = @_;

    my $xml = new XML::Simple;

    my $data = $xml->XMLin($configfilename);

    @instance = grep $_->{'instance'} eq $authZinstance, @{$data->{'authz'}};

    if (scalar @instance != 1) {
	errorExit("Please specify valid and unique RBAC instance");
    } else {
	$rbacInstance = $instance[0];
    }

    $SLCoptions = $rbacInstance ->{'SLCsupport'};

    return {
    rbacbase => $rbacInstance->{rbacbase},
    noDelegationURL => $SLCoptions->{noDelegationURL},
    portalDelegationURL => $SLCoptions->{portalDelegationURL},
    SLCSaddress => $SLCoptions->{SLCSaddress}
    };


}


sub invokeWebService {
    my ($endpointInWSDL,$namespace,$methodname,$arghash) = @_;

    use SOAP::Lite;

    my $soap = SOAP::Lite->new( proxy => $endpointInWSDL);
    $soap->default_ns($namespace);

    @args = ();
    foreach $k (keys %{$arghash}) {
	push @args, SOAP::Data->name($k)->value( $arghash->{$k});
    }

    my $result = $soap->call( $methodname, @args );

    die $result->fault->{ faultstring } if ($result->fault);

    return $result->freeform; # i.e. a hash
}


sub errorExitWithSID { 
    my ($eppn, $sid, $noDelegationURL, $format) = @_;
    print $cgi->header();
    print $cgi->start_html(
	-title=>"Delegation not successful");

    print sprintf("Error: " . $format, @_);

    print "<br/>\n";
    print "Delegation did not succeed. However, you can also work without a certificate now.<br/>\n";
    print "<form method=\"post\" action=\"" . $noDelegationURL . "\">\n";
    print "<input type=\"hidden\" name=\"ePPNplusSID\" value=\"" . $eppn ."|". $sid . "\" />\n";
    print "<input type=\"submit\" value=\"Work without a Certificate\"/>\n";
    print "</form><br/>\n";

    print $cgi->end_html();
    exit(1);
}


# errorExit()
# Dump HTML error and exit
sub errorExit
{
    my $format = shift;
    print $cgi->header();
    print sprintf("Error: " . $format, @_);
    exit(1);
}

